# frozen_string_literal: true

# Debug
require 'pry'
# Generar UUIDs
require 'securerandom'
# Traer resultados de la base de datos
require 'sequel'
require 'sqlite3'
require 'json'
# Limpieza de contenido
require 'loofah'
require 'rails/html/scrubbers'
require 'rails/html/sanitizer'
require 'reverse_markdown'
# Descargar archivos
require 'faraday'

class Wordpress
  attr_reader :site, :prefix, :limit, :url

  def initialize(site:, url:, prefix: 'wp_', limit: 10)
    @site = site
    @prefix = prefix.freeze
    @limit = limit.freeze
    @url = url.freeze

    # Conectarse a la base de datos
    @wp = Sequel.sqlite(File.join(site.source, '_data', 'wordpress', 'post.sqlite3'))
    # Las funciones de JSON usan mucha CPU, vamos a traer de a pocos
    # registros.
    @wp.extension :pagination
  end

  def download(file)
    dest = 'wp-content/uploads/' + file
    full = File.join(site.source, dest)

    return dest if File.exist? full

    Jekyll.logger.info "Downloading #{dest}"

    FileUtils.mkdir_p File.dirname(full)

    File.open(full, 'w') do |f|
      Faraday.get(url + '/' + dest) do |req|
        req.options.on_data = Proc.new do |chunk, _|
          f.write chunk
        end
      end
    end

    dest
  end

  # Obtiene todos los posts opcionalmente filtrando por tipo de post.
  # No es la forma oficial de Sequel pero no tenemos tiempo de
  # aprenderla específicamente y además tenemos las opciones en formato
  # JSON que no estarían soportadas.
  def posts(layout: nil)
    query  = post_query.dup
    query += " where post_type = '#{layout}'" if layout
    query += ' group by posts.ID'

    @wp[query].each_page(limit)
  end

  def meta(id:)
    @wp[meta_query(id: id)].to_a
  end

  private

  # Obtener todos los posts, json_objectagg requiere mariadb 10.5
  def post_query
    @post_query ||= <<~EOQ
      select ID as id,
        post_title as title,
        post_name as slug,
        post_type as layout,
        strftime('%Y-%m-%d', post_date) as date,
        post_status as status,
        post_content as content,
        json_group_object(meta_key, meta_value) as data
      from #{prefix}posts as posts
      left join #{prefix}postmeta as frontmatter
      on posts.ID = frontmatter.post_id
    EOQ
  end

  def meta_query(id:)
    <<~EOQ
      SELECT
        terms.name AS `name`,
        ttax.taxonomy AS `type`,
        ttax.parent AS `parent`,
        ttax.term_id AS `id`
      FROM
        #{prefix}terms AS `terms`,
        #{prefix}term_relationships AS `trels`,
        #{prefix}term_taxonomy AS `ttax`
      WHERE
        trels.object_id = '#{id}' AND
        trels.term_taxonomy_id = ttax.term_taxonomy_id AND
        terms.term_id = ttax.term_id
    EOQ
  end
end

# Antes de generar el sitio vamos a leer todos los artículos desde la
# base de datos y generarlos localmente.
Jekyll::Hooks.register :site, :post_read do |site|
  wp = Wordpress.new(site: site,
                     url: site.config.dig('wordpress', 'url'),
                     prefix: site.config.dig('wordpress', 'prefix'))

  collection = site.collections['posts']
  ascii_re = Regexp.new("\P{ASCII}").freeze
  sanitizer = Rails::Html::SafeListSanitizer.new

  # Traer todas las imágenes cargadas y descargarlas
  attachments = wp.posts(layout: 'attachment').map do |page|
    page.map do |attachment|
      attachment[:data] = JSON.parse(attachment[:data]) unless attachment[:data].nil?
      file = attachment.dig(:data, '_wp_attached_file')

      next unless file

      dest = wp.download(file)

      # Tener un mapa de IDs y archivos destino
      [ attachment[:id], dest ]
    end
  end.compact.flatten(1).to_h

  %w[post page].each do |type|
    wp.posts(layout: type).each do |page|
      page.each do |post|
        # Convertir los datos extra en un Hash
        post[:data] = JSON.parse(post[:data]) unless post[:data].nil?
        post[:slug] = Jekyll::Utils.slugify(post[:title], mode: 'latin') if post[:slug].empty?
        post[:meta] = wp.meta id: post[:id]

        path = File.join(site.source, '_posts', post.slice(:date, :slug).values.join('-') + '.markdown')

        if File.exist? path
          Jekyll.logger.info "#{path} ya fue migrado, actualizando"

          doc = site.documents.find do |d|
            d['id'] == post[:id]
          end
        else
          # Crear un post nuevo y agregarlo a la colección
          collection.docs << doc = Jekyll::Document.new(path, site: site, collection: collection)
          doc.data['uuid'] = SecureRandom.uuid
        end

        thumbnail = post.dig(:data, '_thumbnail_id')&.to_i

        doc.data['layout'] = type
        doc.data['title'] = post[:title]
        doc.data['draft'] = post[:status] != 'publish'
        doc.data['id'] = post[:id]
        doc.data['date'] = Jekyll::Utils.parse_date(post[:date])
        doc.data['tags'] = post[:meta].select { |k| k[:type] == 'post_tag' }.map { |k| k[:name] }
        doc.data['categories'] = post[:meta].select { |k| k[:type] == 'category' }.map { |k| k[:name] }
        doc.data['image'] = attachments[thumbnail] if thumbnail

        doc.content = ReverseMarkdown.convert(sanitizer.sanitize(post[:content]))
        doc.save
      end
    end
  end
end
